/******************************* Copyright notice *******************************
Copyright (C)2002 by Andrei Loskutov <loskutov@gmx.de>. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
********************************************************************************/

package adiag.connection;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Hashtable;
import java.util.StringTokenizer;

import javax.swing.JOptionPane;

//import adiag.gui.App;
import adiag.model.Const;
import adiag.model.Version;
import adiag.model.VersionComparator;
//import adiag.view.Toolz;

/**
 * The connection to Perforce-prozess as encapsulated objekt.
 * @author Andrei Loskutov
 */
public abstract class Connector {
    /** instance for sorting */
    private static VersionComparator comp = new VersionComparator();

    /** which branches we know? 
     * key is path as String, value is ArrayList with versions
     */
    protected Hashtable br_table;
    /** the Debug-Messages-output-list */
    protected StringWriter sout;
    /** the Debug-Messages-output-list */
    protected PrintWriter out;
    /** current conf instance */
    protected Configuration conf; 
    /** Instance of Runtime */
    protected Runtime rt;
    /** file Instance for the 1:1 file output of input */
    protected File filelog;    
    /** file name for the 1:1 file output of input */
    protected static String file_out_name = "tp4branchview.log";
    /** directory, for the 1:1 file output of input, default is "user.home" from JDK - Properties */
    protected static String file_out_dir = System.getProperty("user.home");
      
    /**
     * Default Constructor, call init()
     */
    protected Connector() {
        super();
		init();          
    }
    
	/**
	 * will be called from constructor
	 */
	protected void init() {
		br_table = new Hashtable(13);
		sout = new StringWriter(1000);
		out = new PrintWriter( sout );
	}
    
    /**
     * Constructor 
     * @param app current GUI, for error feedback etc
     */
    public Connector(Configuration conf) {
        this();  
		this.conf = conf;		    
    }
    
    
    /**
     * close connector and all probably opened streams
     * @throws IOException if something goes wrong
     */
    public abstract void close() throws IOException ;
    
    /**
     * dispose all connector resources and close all probably opened streams
     * @throws IOException if something goes wrong
     */
    public void dispose() throws IOException{
        close();
        filelog = null;
        if(sout != null){
            sout.close();
        }
        if(out != null){
            out.close();
        }
        sout = null;
        out = null;
    }
        
    /**
     * convert all data / error messages from currrent data stream to String
     * @return convert all data / error messages from currrent data stream to String
     */
    public String dataToString() {
        if(out == null){
            return "";
        }
        out.flush();
        return sout.toString();
    }
    
    /**
     * See return description
     * @return Array with all "root"- revisions from all branches, or empty aray
     * @throws Exception if something goes wrong
     */
    public Version[] getData() throws Exception{
        String path = conf.getPath();
        
        if(path != null){            
            String [] paths = null;
            if(conf.isDiscoverAllBranches()){
                paths = getUnknownBranches(false);
            }
            if( paths == null || paths.length == 0){            	
                paths = new String []{path};                
            } 
            System.out.println("try to receive data: "+ makeLine(paths, " "));                  
            return getP4Data(paths);            
        } else {
            return new Version[0];
        }
    }

    /**
     * See return description
     * @param words strings to join
     * @param seperator between
     * @return one big line with given strings separated with given separator
     */
     public static String makeLine(String[] words, String seperator) {
         StringBuffer sb = new StringBuffer();
         for (int i = 0; i < words.length; i++) {
             sb.append(words[i]);
             if (i < words.length - 1)
                 sb.append(seperator);
         }
         return sb.toString();
     }
    
	/**	
     * See return description
	 * @return File the instance of current Log-file
	 */
	public final File getLogFile(){
		if(filelog == null){		    
			return new File(file_out_dir, file_out_name);
		} else {
			return filelog;
		}
	}   
	
	/**
     * See return description
	 * @return String the path to Diff - tool
	 */
	protected final String getDiffPath(){
		return conf.getDiffPath();
	}   		
	
	
	/**
	 * Method setLogFile set the instance of Log-file
	 * @param f the instance of Log-file
	 */
	public final void setLogFile(File f){		
		filelog = f;
	} 
      
       
    /**
     * Return array with "root"-Versions (each one for different branch)
     * <pre>
     *  //give not all needed information
     *	FileEntry fe = new FileEntry(env, "//apis/entw/src/batch/batch_i/bat_glob.h");
     *	// ist nicht fertig implementirt:
     *	Vector v = FileEntry.getFileLog(env, "//apis/entw/src/batch/batch_i/bat_glob.h");
     *	System.out.println("filelog hat " +v.size() +" Eintrge");
     *	for (int i = 0; i < v.size(); i++){
     *		FileEntry f = (FileEntry)v.get(i);
     *		System.out.println(f.toXML());
     *	}
     * </pre>
     * @param apath array with paths of files to request it from perforce
     * @return Array with "root"-Versions (each one for different branch), or an empty array
     * @throws Exception if something goes wrong
     */
    protected Version[] getP4Data(String [] apath) throws Exception {               
        //System.out.println("received paths: " + Toolz.makeLine(apath, " : "));
        /*if(apath.length > 10){
            if(!app.yesno( 
                "You are about to send request over " + apath.length 
                + " files at one time."+
                " This may cause some troubles. Are you sure ???")){               //\nAre you sure ??? 
                return getVersions();
            }
        }   */      
        // create or update branch data
        readP4Out(apath);
        
        ArrayList p = null;
        Version [] ver = null;
        ArrayList pend = new ArrayList(br_table.keySet());
        Version tmp;
        ArrayList need_to_discover = new ArrayList();
        for (int i = 0; i < pend.size(); i++){
            p = (ArrayList)br_table.get(pend.get(i));                        
            // check for new data, set some flags and update todo list
            for (int j = 0; j < p.size(); j++){                
                tmp = (Version)p.get(j);
                // set flag if we receive data directly from perforce
          		tmp.setReceived(tmp.isReceived() || count(apath, tmp.getPath())>0 );          		
          		// update todo list if needed
                if(!tmp.isReceived() && conf.isDiscoverAllBranches() && !tmp.isChangeNumOk()){
          			if(!need_to_discover.contains(tmp.getPath())){
          				need_to_discover.add(tmp.getPath());
          			}
          		}
            }
            ver = sort(p);
            p.clear();
            // update parent connections
            for (int j = 1; j < ver.length; j++){
                ver[j-1].setNext(ver[j]);
                ver[j].setPrev(ver[j-1]);
                p.add(ver[j-1]);
            }
            p.add(ver[ver.length-1]);
            br_table.put(pend.get(i), p);
        }
        // recursive call if todo isn't empty
        if( !need_to_discover.isEmpty() && !conf.isOffline() ){   	
            return getP4Data((String[])need_to_discover.toArray(new String[need_to_discover.size()]));
        }
        
		return getRootVersions();
    }

    /**
     * See return description
     * @param v Version[] array, not null
     * @return Version[] the same list, sorted
     */
    public static Version[] sort(Version[] v) {
        Arrays.sort(v, comp);
        return v;
    }


    /**
     * See return description
     * @param v list with versions, not null
     * @return Version[] sorted array
     */
    public static Version[] sort(ArrayList v) {        
        return sort((Version[]) v.toArray(new Version[v.size()]));
    }

	/**
	 * @return Version[] array with root versions or empty array
	 * @throws IOException if something goes wrong
	 */
	protected Version[] getRootVersions() throws IOException {
		ArrayList c = new ArrayList(br_table.values());
		ArrayList al = new ArrayList();
		for (int i = 0; i < c.size(); i++){
		    // search and return really(!!!) first element
		    ArrayList t = (ArrayList)c.get(i);
		    if(t.size()>0){
		        al.add( getFirst( (Version)t.get(0) ));
		    }
		}					
		return (Version[])al.toArray(new Version[al.size()]);
	}
        
    /**
     * See return description
     * @param list java.util.ArrayList entries are Version objects
     * @return Version the smallest Version (with the smallest change number)
     * from given list or null, if list is empty
     */
    public static Version getFirst(ArrayList list) {
        if (list.size() == 0) {
            return null;
        }
        ArrayList al = new ArrayList();
        Object v;
        // make a copy from list
        for (int i = 0; i < list.size(); i++) {
            v = list.get(i);
            if (v != null) {
                al.add(v);
            }
        }
        if (al.size() == 0) {
            return null;
        }
        Collections.sort(al, comp);
        return (Version) al.get(0);
    }  

    
    /**
     * See return description
     * @param v Version any one version from branch
     * @return Version the first Version (with the smallest version number)
     * from given branch
     */
    public static Version getFirst(Version v) {
        Version p = v;
        while ((v = v.getPrev()) != null) {
            p = v;
        }
        return p;
    }
     
	/**
     * read native P4 Log 
     * @param apath initial paths for requested files/directories
     * @throws Exception if something goes wrong
     */
    protected abstract void readP4Out(String [] apath)  throws Exception;


    /** See return description
    * @param sa String array
    * @param value string to count
    * @return count number of occurrences of given string in given array
    */
    public static int count(String[] sa, String value) {
        int c = 0;
        int length = sa.length;
        for (int i = 0; i < length; i++) {
            if (sa[i].equals(value))
                c++;
        }
        return c;
    }

/*
 * log  example for file //depot/perforce/testroot/itest/test.txt:
 * 
info: //depot/perforce/testroot/itest/test.txt
info1: #4 change 107 integrate on 2002/05/27 by Andy_the_Great@andy (text)
info1: 
info1: 	merge test
info2: merge from //depot/perforce/testroot/prod/test.txt#1,#3
info1: #3 change 93 edit on 2002/05/25 by Andy_the_Great@andy (text)
info1: 
info1: 	noch ein edit...
info2: copy into //depot/perforce/testroot/entw/test.txt#6
info2: ignored by //depot/perforce/testroot/kdtest/test.txt#4
info1: #2 change 67 edit on 2002/05/14 by Andy_the_Great@andy (text)
info1: 
info1: 	bearbeitet
info2: branch into //depot/perforce/testroot/entw/test.txt#1
info1: #1 change 66 branch on 2002/05/14 by Andy_the_Great@andy (text)
info1: 
info1: 	kdtest zum itest
info2: branch from //depot/perforce/testroot/kdtest/test.txt#1,#2
info: //depot/perforce/testroot/kdtest/test.txt
info1: #2 change 65 edit on 2002/05/14 by Andy_the_Great@andy (text)
info1: 
info1: 	bearbeitet
info2: branch into //depot/perforce/testroot/itest/test.txt#1
info1: #1 change 64 branch on 2002/05/14 by Andy_the_Great@andy (text)
info1: 
info1: 	branch zum kdtest vom prod
info2: branch from //depot/perforce/testroot/prod/test.txt#1,#2
info: //depot/perforce/testroot/prod/test.txt
info1: #2 change 63 edit on 2002/05/14 by Andy_the_Great@andy (text)
info1: 
info1: 	bearbeitet
info2: branch into //depot/perforce/testroot/entw/test.txt#4
info2: branch into //depot/perforce/testroot/kdtest/test.txt#1
info1: #1 change 62 add on 2002/05/14 by Andy_the_Great@andy (text)
info1: 
info1: 	Test erstellt im prod
exit: 0
*/

    /**
     * parse data from LineReader to the model - Objekts and call close() after work
     * @param apath first path in log file
     * @param in LineReader with raw data
     * @throws Exception if something goes wrong
     */
	protected void parseP4FileLog(String apath, LineReader in) throws Exception {	
		String l = null, vnum, temp;
		int changenum = 0;
		Version v = null;
		String[] words = null;
		boolean from_to = false;
		/* We assume that the p4 command was called with -s which sends all 
         * messages to standard out pre-pended with a string that indicates 
         * what kind of messsage it is: error,warning,text,info,exit
		 */  
		while (null != (l = in.readLine())) {
            out.println(l);
	        if(l.indexOf(":")>=0){
	            if(!l.startsWith("info")){
	            	continue;
	            } else {    		            
		            // is it path line? 
		            temp = l.substring(l.indexOf(":") + 1);
		            if(l.startsWith("info:")){
		                apath = temp.trim();
		            	continue;
		            } else {
		            	from_to = l.startsWith("info2:");
		            	l = temp;    		            	
		            }	            
	            }		            		            		        	
	        } else {
                continue;
            }
	        // parse version nummer, create version, we have the change - Line
	        // we test for " #", because p4 make *one* space character before '#'
            // only on changes, 
	        // and it may be simple comment line with '#' at start;
            // P4 make then one space character *and* one tab stop before.
	        if (l.startsWith(" #") && l.indexOf("change") > 1) {
	            // we don't need trim because getWords do it for us            
	            // read number from: 
	            // #3 change 903 edit on 2001/11/09 by loskut@loskut-wa114387 (ktext)
	            // #6 change 1667 integrate on 2002/05/27 by schroed@schroed-whaj3103 (ktext)
	            words = split(l);
	            // words contain now all information
	            vnum = words[0].substring(1, words[0].length());
	            int vnumi = intValue(vnum);
	            changenum = intValue(words[2]);
	            v = getVersion(apath, vnumi, true);
	            // only if change unknown!!!
                if(!v.isChangeNumOk()){
	                v.setChange(changenum);
	            }
	            v.setChangeType(words[3]);	                
	            v.setDate(words[5]);
	            v.setUser(words[7]);		            
	        } else {// Branch/Edit/etc informations or comment
	            l = l.trim();     		            
	            int startIdx = -1; 
                // comment!
	            if(!from_to || (startIdx = Const.getStartLineCode(l)) < 0 ){
	                if (l.length() > 0) {                        
	                    if(v!=null) { 							
	                    	v.addDescriptionLine(l);
	                    }
	                }
	                continue;
	            }		            
	            int command_length = Const.getStartLineCodeLength(startIdx);
	            int start_change = l.indexOf("#");
	            // branch from //depot/perforce/testroot/kdtest/test.txt#1,#2
	            String b_path = l.substring(command_length, start_change);
	            int v_idx = l.lastIndexOf("#");
	            int end_idx = l.length();
	
	            String b_vnum = l.substring(v_idx + 1, end_idx);
	
	            Version b_v = getVersion(b_path, intValue(b_vnum), true);
	            // set branch to/from connections
	            if (startIdx < Const.FROM_NUMBER) {
	                b_v.addNextBranch(v);		                
	                v.addPrevBranch(b_v, Const.LINES[startIdx]);
	            } else {
	                b_v.addPrevBranch(v, Const.LINES[startIdx - Const.FROM_NUMBER]);
	                v.addNextBranch(b_v);
	            }
	        } // end else
	    } // end while
	    close();
        System.runFinalization();
	    System.gc();	
	}

    /** Split one String into many, as separator is the " "
    * @param s string to split
    * @return array with substrings, at least with the same string
    */
    public static String[] split(String s) {
        return split(s, " ");
    }
    
    /** Split one String into many, as separator is the " "
    * @param s string to split
    * @param delimiters any character in a string is separator
    * @return array with substrings, at least with the same string
    */
    public static String[] split(String s, String delimiters) {
        StringTokenizer st = new StringTokenizer(s, delimiters);
        String[] arr = new String[st.countTokens()];
        for (int i = 0; i < arr.length; i++) {
            arr[i] = st.nextToken();
        }
        return arr;
    }
    
    /**
     * See return description
     * @return int value of given string as int or 0, if it is not a int
     * @param s java.lang.String
     */
    public static int intValue(String s) {
        int i = 0;
        try {
            i = Integer.parseInt(s);
        } catch (NumberFormatException e) {
            System.out.println(e);
        }
        return i;
    }

	/**
	 * Compare two versions, but as default does nothing and must be overriden!
	 * @param a first Version
	 * @param b second Version
     * @throws Exception if something goes wrong
	 */
	public void diff(Version a, Version b) throws Exception {
		JOptionPane.showMessageDialog(
			conf.getJComponent(), 
			"This feature is not available now.", 
			"Information", 
			JOptionPane.INFORMATION_MESSAGE);	
	}

	/**
	 * See return description
	 * @param report_even_if_not_dicover_all if true, return all branches even 
     * if only one at time is allowed to be discover (by settings)
	 * @return String[] all paths of branches, if thay haven't full information about 
     * change numbers or if they been never directly examined
	 */
	public String[] getUnknownBranches(boolean report_even_if_not_dicover_all){
		ArrayList p = null;        
		ArrayList pend = new ArrayList(br_table.keySet());
        Version tmp = null;
        ArrayList need_to_discover = new ArrayList();
        boolean discover_all = conf.isDiscoverAllBranches() || report_even_if_not_dicover_all;
        for (int i = 0; i < pend.size(); i++){
            p = (ArrayList)br_table.get(pend.get(i));                                    
            for (int j = 0; j < p.size(); j++){                
          		tmp = (Version)p.get(j);
          		if(!tmp.isChangeNumOk() || (!tmp.isReceived() && discover_all)){
          			if(!need_to_discover.contains(tmp.getPath())){          				
                        need_to_discover.add(tmp.getPath());
          			}
          		}
            }            
        }	                                
		return (String[])need_to_discover.toArray(new String[need_to_discover.size()]);
	}    
            
    /**
     * Search version from branch table and if not found, create one (depend on parameter)
     * @return founded / created version or null, of nothing found and 'create' flag is false
     * @param path version path
     * @param num version number
     * @param create true, to create new version, if nothing found
     */
    protected Version getVersion(String path, int num, boolean create) {
        ArrayList ht = (ArrayList)br_table.get(path);
        if(ht == null){ // branch don't exist yet
            ht = new ArrayList();
            br_table.put(path, ht);
        }        
        Version v = getForNummer(ht, num);
        if(v == null && create){ // version don't exist yet!
            v = new Version();
            v.setVersNumber(num);
            v.setPath(path);
            ht.add(v);
        }   
        return v;
    }       
    
    /**
     * See return description
     * @param c ArrayList, entries are Version objects
     * @param num the version number
     * @return Version with given version number,or null, if not found
     */
    public static Version getForNummer(ArrayList c, int num) {
        if (c == null || c.size() == 0) {
            return null;
        }
        Version v = null;
        for (int i = 0; i < c.size(); i++) {
            v = (Version) c.get(i);
            if (num == v.getVersNumber()) {
                return v;
            }
        }        
        return null;
    }    
    
}
